@@ -2,10 +2,9 @@
 import math
 import tkinter as tk
 from functools import partial
-from tkinter import filedialog, ttk
-from typing import Mapping, Optional
+from tkinter import ttk
+from typing import Mapping, MutableMapping, Optional
 
-import dask.dataframe as dd
 import numpy as np
 
 from ..core.parameters import (
@@ -13,22 +12,24 @@ from ..core.parameters import (
     BaseParam,
     BooleanParam,
     ChoiceParam,
+    ColumnOrStringParam,
     FileArrayParam,
     FileParam,
     FileSaveParam,
-    FloatParam,
-    IntegerParam,
     MultiParam,
     SimpleParam,
+    TabularMultiParam,
     TextParam,
 )
 from ..core.plugins import BasePlugin
-from ..utils.dask import crop_dataframe
-
-UNICODE_CHECK = "\u2714"
-UNICODE_UNCHECK = "\u2717"
-UNICODE_CROSS = "\u2715"
-UNICODE_PLUS = "\u2795"
+from .widgets import (
+    BooleanCheckbox,
+    add_button,
+    ask_open_filename,
+    ask_open_filenames,
+    ask_saveas_filename,
+    delete_button,
+)
 
 
 def is_nan(v):
@@ -50,35 +51,41 @@ class ParameterWrapper:
         delete_callback=None,
         level=0,
     ):
+        self.tk_parent = tk_parent
         self.parameter = parameter
         self.callback = callback
         self.button = None
         self.level = level
         self.subwrapper_buttons: list[tk.Button] = []
+        self.column_labels: list[tk.Label] = []
 
         self.subwrappers: Mapping[BaseParam, ParameterWrapper] = {}
 
         self.var: Optional[tk.Variable] = None
         self.entry: Optional[tk.Widget] = None
         self.label: Optional[tk.Widget] = None
+        self.row_labels: list[tk.Widget] = []
 
         if isinstance(parameter, ArrayParam):
             self.label = None
         else:
-            self.label = ttk.Label(tk_parent, text=parameter.label)
+            self.label = tk.Label(tk_parent, text=parameter.label)
 
         if isinstance(parameter, ChoiceParam):
             self.var = tk.StringVar(tk_parent, value=parameter.value)
             self.entry = ttk.Combobox(tk_parent, textvariable=self.var)
-            self.entry["values"] = parameter.choices
-            self.entry.state(["readonly"])
+            self.entry["values"] = parameter.choices or [""]
+            if isinstance(parameter, ColumnOrStringParam):
+                self.entry.bind("<Key>", self.combobox_set)
+                self.entry["state"] = "normal"
+            else:
+                self.entry["state"] = "readonly"
         elif isinstance(parameter, BooleanParam):
-            self.entry = tk.Button(tk_parent, width=2, command=self.toggle_checkbox_callback)
-            self.set_checkbox_value(parameter.value)
+            self.entry = BooleanCheckbox(tk_parent, command=self.toggle_checkbox_callback)
+            self.set_checkbox_value()
         elif isinstance(parameter, FileParam):
-            self.var = tk.StringVar(tk_parent, value=parameter.value)
-            self.entry = ttk.Entry(tk_parent, textvariable=self.var)
-            self.entry.state(["readonly"])
+            self.entry = tk.Label(tk_parent, text=parameter.value)
+            self.button = tk.Button(tk_parent, text="Select", width=3, command=self.change_file_callback)
         elif isinstance(parameter, TextParam):
             # tk.Text widget doesn't have a variable, for whatever reason,
             # so we use a different method!
@@ -86,23 +93,19 @@ class ParameterWrapper:
             self.entry = tk.Text(tk_parent, height=10)
             self.entry.insert("1.0", parameter.value)
             if parameter.read_only:
-                self.entry["state"] = "disabled"
+                self.entry["state"] = tk.DISABLED
             else:
                 self.entry.bind("<<Modified>>", self.widget_modified_callback)
         elif isinstance(parameter, SimpleParam):
-            if isinstance(parameter, FloatParam):
-                self.var = tk.DoubleVar(tk_parent, value=parameter.value)
-            elif isinstance(parameter, IntegerParam):
-                self.var = tk.IntVar(tk_parent, value=parameter.value)
-            else:
-                self.var = tk.StringVar(tk_parent, value=parameter.value)
-
-            self.entry = ttk.Entry(tk_parent, textvariable=self.var)
+            self.var = tk.StringVar(tk_parent, value=parameter.value)
+            self.entry = tk.Entry(tk_parent, textvariable=self.var)
             if parameter.read_only:
-                self.entry.state(["readonly"])
+                self.entry["state"] = tk.DISABLED
 
-        elif isinstance(parameter, ArrayParam) and self.level == 0:
-            self.entry = ttk.Frame(tk_parent)
+        elif (
+            isinstance(parameter, ArrayParam) and self.level == 0 and not isinstance(parameter.param, TabularMultiParam)
+        ):
+            self.entry = tk.Frame(tk_parent)
             self.entry.columnconfigure(0, weight=1)
             drc = self.delete_row_callback if not parameter.read_only else None
             self.update_subwrappers_framed(parameter.params, drc)
@@ -114,24 +117,16 @@ class ParameterWrapper:
                 )
                 self.button.grid(row=len(parameter.params) + 1)
 
-        elif isinstance(parameter, ArrayParam) and self.level < 3:
+        elif isinstance(parameter, ArrayParam) and (self.level < 3 or isinstance(parameter.param, TabularMultiParam)):
             label_frame_label = tk.Frame(tk_parent)
             tk.Label(label_frame_label, text=parameter.label).grid(row=0, column=0, padx=5)
             self.entry = tk.LabelFrame(tk_parent, labelwidget=label_frame_label, padx=10, pady=5)
-            self.button = tk.Button(
-                label_frame_label,
-                text=UNICODE_PLUS,
-                width=2,
-                command=self.add_row_callback,
-            )
-            self.button.grid(row=0, column=1, padx=10)
+            if not parameter.read_only:
+                self.button = add_button(label_frame_label, command=self.add_row_callback)
+                self.button.grid(row=0, column=1, padx=10)
 
             drc = self.delete_row_callback if not parameter.read_only else None
             if isinstance(parameter.param, MultiParam):
-                for n, pp in enumerate(parameter.param.values()):
-                    tk.Label(self.entry, text=pp.label).grid(row=0, column=n, sticky=tk.EW, padx=10)
-                    self.entry.columnconfigure(n, weight=1)
-
                 self.update_subwrappers_tabular(parameter.params, drc)
             else:
                 self.entry.columnconfigure(0, weight=0)
@@ -145,16 +140,14 @@ class ParameterWrapper:
                 self.update_subwrappers(parameter.params, drc)
 
         elif isinstance(parameter, (ArrayParam, MultiParam)):
-            self.entry = ttk.Frame(tk_parent)
+            self.entry = tk.Frame(tk_parent)
             self.entry.columnconfigure(0, weight=0)
             self.entry.columnconfigure(1, weight=0)
             self.entry.columnconfigure(2, weight=1)
             if isinstance(parameter, ArrayParam):
                 drc = self.delete_row_callback if not parameter.read_only else None
                 self.update_subwrappers(parameter.params, drc)
-                self.button = tk.Button(
-                    tk_parent, text=UNICODE_PLUS, width=2, command=self.add_row_callback
-                )
+                self.button = add_button(tk_parent, command=self.add_row_callback)
             else:
                 self.update_subwrappers(parameter.params.values(), None)
         else:
@@ -164,46 +157,64 @@ class ParameterWrapper:
             self.var.trace("w", self.value_changed_callback)
 
         # XXX hang on, what if it's an array in an array?
-        if delete_callback and not self.button:
-            self.button = tk.Button(
-                tk_parent,
-                text=UNICODE_CROSS,
-                width=2,
-                command=lambda: delete_callback(self),
-            )
+        if not parameter.read_only and delete_callback and not self.button:
+            self.button = delete_button(tk_parent, command=lambda: delete_callback(self))
 
         if not isinstance(parameter, BooleanParam):
             self.entry.grid(sticky=tk.EW, padx=10, pady=5)
 
     def update(self):
-        if isinstance(self.parameter, ArrayParam) and self.level == 0:
+        if self.parameter.hide:
+            self.entry["fg"] = self.entry["bg"]
+        else:
+            self.entry["fg"] = None
+
+        if (
+            isinstance(self.parameter, ArrayParam)
+            and self.level == 0
+            and not isinstance(self.parameter.param, TabularMultiParam)
+        ):
             self.update_subwrappers_framed(self.parameter.params, self.delete_row_callback)
             if self.button:
                 self.button.grid(row=len(self.parameter.params) + 1, padx=10)
-        elif (
-            isinstance(self.parameter, ArrayParam)
-            and isinstance(self.parameter.param, MultiParam)
-            and self.level < 3
-        ):
-            self.update_subwrappers_tabular(self.parameter.params, self.delete_row_callback)
+        elif isinstance(self.parameter, ArrayParam) and isinstance(self.parameter.param, MultiParam) and self.level < 3:
+            self.update_subwrappers_tabular(
+                self.parameter.params,
+                self.delete_row_callback if not self.parameter.read_only else None,
+            )
         elif isinstance(self.parameter, ArrayParam):
-            self.update_subwrappers(self.parameter.params, self.delete_row_callback)
+            self.update_subwrappers(
+                self.parameter.params,
+                self.delete_row_callback if not self.parameter.read_only else None,
+            )
             if self.button:
                 self.button["state"] = (
                     tk.DISABLED
-                    if self.parameter.max_size is not None
-                    and len(self.subwrappers) >= self.parameter.max_size
+                    if self.parameter.max_size is not None and len(self.subwrappers) >= self.parameter.max_size
                     else tk.NORMAL
                 )
         elif isinstance(self.parameter, MultiParam):
             self.update_subwrappers(self.parameter.params.values(), None)
+        elif isinstance(self.parameter, ChoiceParam):
+            choices = self.parameter.choices or [""]
+            self.entry["values"] = choices
+            self.var.set(self.parameter.value)
+        elif isinstance(self.parameter, BooleanParam):
+            self.set_checkbox_value()
+        elif isinstance(self.parameter, TextParam):
+            if self.parameter.read_only:
+                self.entry["state"] = "normal"
+            self.entry.replace("1.0", tk.END, self.parameter.value)
+            if self.parameter.read_only:
+                self.entry["state"] = "disabled"
+        elif isinstance(self.parameter, FileParam):
+            self.entry["text"] = self.parameter.value
+        else:
+            self.var.set(self.parameter.value)
 
         if self.label:
             self.label["text"] = self.parameter.label
 
-        if isinstance(self.parameter, ChoiceParam):
-            self.entry["values"] = self.parameter.choices
-
     def cull_subwrappers(self, params):
         params_set = set(params)
         for p, pw in self.subwrappers.items():
@@ -229,8 +240,22 @@ class ParameterWrapper:
     def update_subwrappers_tabular(self, params, delete_row_callback):
         while self.subwrapper_buttons:
             self.subwrapper_buttons.pop().destroy()
+        while self.row_labels:
+            self.row_labels.pop().destroy()
+        while self.column_labels:
+            self.column_labels.pop().destroy()
+
+        for n, pp in enumerate(self.parameter.param.values()):
+            column_label = tk.Label(self.entry, text=pp.label)
+            self.column_labels.append(column_label)
+            column_label.grid(row=0, column=n + 1, sticky=tk.EW, padx=10)
+            self.entry.columnconfigure(n + 1, weight=1)
 
         for n, p in enumerate(params):
+            row_label = tk.Label(self.entry, text=p.label)
+            row_label.grid(row=n + 1, column=0, padx=10)
+            self.row_labels.append(row_label)
+
             subparams = p.params.values()
             for m, pp in enumerate(subparams):
                 if pp in self.subwrappers:
@@ -243,15 +268,10 @@ class ParameterWrapper:
                         delete_row_callback,
                         level=self.level + 1,
                     )
-                self.subwrappers[pp].entry.grid(row=n + 1, column=m, padx=10)
-            if delete_row_callback:
-                button = tk.Button(
-                    self.entry,
-                    text=UNICODE_CROSS,
-                    width=2,
-                    command=partial(delete_row_callback, self, n),
-                )
-                button.grid(row=n + 1, column=len(subparams), padx=10)
+                self.subwrappers[pp].entry.grid(row=n + 1, column=m + 1, padx=10)
+            if not self.parameter.read_only and delete_row_callback:
+                button = delete_button(self.entry, command=partial(delete_row_callback, self, n))
+                button.grid(row=n + 1, column=len(subparams) + 1, padx=10)
                 self.subwrapper_buttons.append(button)
 
         self.cull_subwrappers([pp for p in params for pp in p.params.values()])
@@ -278,19 +298,19 @@ class ParameterWrapper:
                     delete_row_callback,
                     level=self.level + 1,
                 )
-                if delete_row_callback:
-                    button = ttk.Button(
-                        label_frame_label,
-                        text=UNICODE_CROSS,
-                        width=2,
-                        command=partial(_command_drc, p, label_frame),
-                    )
+                if not self.parameter.read_only and delete_row_callback:
+                    button = delete_button(label_frame_label, command=partial(_command_drc, p, label_frame))
                     button.grid(row=0, column=1, padx=10)
 
             self.subwrappers[p].entry.grid(row=n, column=0, padx=10)
 
         self.cull_subwrappers(params)
 
+    def combobox_set(self, event):
+        assert isinstance(self.entry, ttk.Combobox)
+        if self.entry.current() != -1:
+            self.entry.set(event.char)
+
     def set_row(self, row):
         if not self.label:
             self.entry.grid(row=row, column=0, columnspan=3)
@@ -307,11 +327,11 @@ class ParameterWrapper:
 
         if isinstance(self.parameter, FileSaveParam):
             file_types = self.parameter.file_types
-            filename = filedialog.asksaveasfilename(filetypes=file_types)
+            filename = ask_saveas_filename(self.parameter.value, file_types)
             self.parameter.value = filename
         if isinstance(self.parameter, FileArrayParam):
             file_types = self.parameter.file_types
-            filenames = filedialog.askopenfilenames(filetypes=file_types)
+            filenames = ask_open_filenames(file_types)
             self.parameter.add_files(filenames)
         else:
             self.parameter.add_row()
@@ -323,6 +343,7 @@ class ParameterWrapper:
 
     def delete_row_callback(self, parameter_wrapper, row=None):
         assert isinstance(self.parameter, ArrayParam)
+
         if row is not None:
             self.parameter.del_row(row)
         else:
@@ -332,6 +353,13 @@ class ParameterWrapper:
         if self.callback is not None:
             self.callback(self.parameter)
 
+    def change_file_callback(self, *_):
+        file_types = self.parameter.file_types
+        filename = ask_open_filename(file_types)
+        self.parameter.value = filename
+        self.entry["text"] = self.parameter.value
+        self.callback(self.parameter)
+
     def set_value(self, value):
         # self.parameter.value is a property, and some cleaning may occur, so we just
         # check before and after to see if it has changed.
@@ -342,8 +370,17 @@ class ParameterWrapper:
             self.callback(self.parameter)
         return self.parameter.value
 
+    def set_choice(self, choice):
+        self.entry.current(choice)
+        if self.parameter.choice != choice:
+            self.parameter.choice = choice
+            self.callback(self.parameter)
+
     def value_changed_callback(self, *_):
-        self.var.set(self.set_value(self.var.get()))
+        if isinstance(self.parameter, ChoiceParam) and self.entry.current() != -1:
+            self.set_choice(self.entry.current())
+        else:
+            self.var.set(self.set_value(self.var.get()))
 
     def widget_modified_callback(self, *_):
         # only gets called the *first* time a modification happens, unless
@@ -352,18 +389,21 @@ class ParameterWrapper:
         self.entry.edit_modified(False)
         self.set_value(value)
 
-    def set_checkbox_value(self, value):
-        if value:
-            self.entry["text"] = UNICODE_CHECK
-            self.entry["fg"] = "black"
+    def set_checkbox_value(self):
+        if self.parameter.hide:
+            self.entry.set_value(None)
+        elif self.parameter.value:
+            self.entry.set_value(True)
         else:
-            self.entry["text"] = UNICODE_UNCHECK
-            self.entry["fg"] = "grey"
+            self.entry.set_value(False)
 
     def toggle_checkbox_callback(self, *_):
+        if self.parameter.read_only or self.parameter.hide:
+            # XXX warn?
+            return
         value = not self.parameter.value
         self.parameter.value = value
-        self.set_checkbox_value(value)
+        self.set_checkbox_value()
         if self.callback is not None:
             self.callback(self.parameter)
 
@@ -385,27 +425,24 @@ class PluginConfigurator:
         self.plugin = plugin
         self.change_callback = change_callback
 
-        self.frame = ttk.Frame(tk_parent)
-        self.frame.columnconfigure(0, weight=1)
+        self.frame = tk.Frame(tk_parent, highlightcolor="purple", highlightthickness=3)
+        self.frame.columnconfigure(0, weight=0)
+        self.frame.columnconfigure(1, weight=0)
+        self.frame.columnconfigure(2, weight=1)
         self.frame.grid(sticky=tk.NSEW)
 
-        self.wrapper_cache: Mapping[str, ParameterWrapper] = {}
-
-        self.subframe = ttk.Frame(self.frame)
-        self.subframe.columnconfigure(0, weight=0)
-        self.subframe.columnconfigure(1, weight=0)
-        self.subframe.columnconfigure(2, weight=1)
-        self.subframe.grid(row=1, sticky=tk.NSEW)
+        self.wrapper_cache: MutableMapping[str, ParameterWrapper] = {}
 
         self.update()
 
     def change_parameter(self, parameter):
         """Called whenever a parameter gets changed"""
-        self.update()
+        # if self.plugin.update():
+        #    self.update()
         if self.change_callback:
             self.change_callback(self)
 
-    def update(self):
+    def update(self) -> None:
         # If there's only a single parameter it is presented a little differently.
         top_level = 0 if len(self.plugin.parameters) == 1 else 1
 
@@ -415,7 +452,7 @@ class PluginConfigurator:
                 self.wrapper_cache[key].update()
             else:
                 self.wrapper_cache[key] = ParameterWrapper(
-                    self.subframe, parameter, self.change_parameter, level=top_level
+                    self.frame, parameter, self.change_parameter, level=top_level
                 )
             self.wrapper_cache[key].set_row(n + 1)
 
@@ -424,73 +461,3 @@ class PluginConfigurator:
             if key not in self.plugin.parameters:
                 wrapper.destroy()
                 del self.wrapper_cache[key]
-
-
-class DataFramePreview:
-    """Provides a visual preview of a Dask dataframe arranged as a table."""
-
-    # XXX uses a treeview, which seemed like a good match but actually a grid-layout
-    # of custom styled labels might work better.
-
-    def __init__(self, tk_parent, ddf: Optional[dd.DataFrame] = None):
-        self.frame = ttk.Frame(tk_parent)
-        self.label = ttk.Label(self.frame, text="DataFrame Preview")
-
-        self.treeview = ttk.Treeview(self.frame, selectmode=tk.NONE)
-
-        self.scrollbar_x = ttk.Scrollbar(
-            self.frame, orient=tk.HORIZONTAL, command=self.treeview.xview
-        )
-        self.scrollbar_y = ttk.Scrollbar(
-            self.frame, orient=tk.VERTICAL, command=self.treeview.yview
-        )
-        self.treeview.configure(xscrollcommand=self.scrollbar_x.set)
-        self.treeview.configure(yscrollcommand=self.scrollbar_y.set)
-
-        self.frame.grid(sticky=tk.NSEW)
-
-        self.frame.columnconfigure(0, weight=1)
-        self.frame.rowconfigure(1, weight=1)
-        self.label.grid(row=0, columnspan=2)
-        self.treeview.grid(row=1, column=0, sticky=tk.NSEW)
-        self.scrollbar_x.grid(row=2, column=0, sticky=tk.EW)
-        self.scrollbar_y.grid(row=1, column=1, stick=tk.NS)
-
-        if ddf is not None:
-            self.update(ddf)
-
-    def update(self, ddf: dd.DataFrame):
-        if len(ddf) > 1000:
-            self.label["text"] = f"DataFrame Preview (1000 rows out of {len(ddf)})"
-            ddf = crop_dataframe(ddf, 1000)
-        else:
-            self.label["text"] = f"DataFrame Preview {len(ddf)} rows"
-
-        # XXX could handle multiindex columns more elegantly than this
-        # (but maybe not in a ttk.Treeview)
-        column_names = [".".join(c) if isinstance(c, tuple) else c for c in ddf.columns]
-        column_types = [dt.kind for dt in ddf.dtypes]
-
-        self.treeview["columns"] = column_names
-
-        for n, cn in enumerate(column_names):
-            self.treeview.heading(n, text=cn)
-
-        for n, ct in enumerate(column_types):
-            # XXX it'd be nicer if we could do "real" decimal point alignment
-            anchor = tk.E if ct in ("i", "f") else tk.W
-            # XXX type signature appears to be wrong, or at least overly restrictive.
-            # I think I'm going to replace treeview anyway so I'm ignoring it for now.
-            self.treeview.column(
-                n, anchor=anchor, width=100, minwidth=100, stretch=tk.YES
-            )  # type: ignore
-
-        for row in self.treeview.get_children():
-            self.treeview.delete(row)
-
-        for n, (index, *values) in enumerate(ddf.itertuples()):
-            values = ["—" if is_nan(v) else str(v) for v in values]
-            self.treeview.insert("", n, text=index, values=values)
-
-    def destroy(self):
-        self.frame.destroy()
