@@ -96,6 +96,8 @@ class ConfiguratorWrapper:
                 self.show_preview_subframe()
 
     def show_config_subframe(self):
+        if self.config_subframe:
+            self.config_subframe.destroy()
         if self.config_canvas:
             self.config_canvas.destroy()
         self.config_canvas = tk.Canvas(self.subframe)
@@ -193,7 +195,7 @@ class ConfiguratorWrapper:
         else:
             try:
                 df = concat_dataframes(self.node.result)
-                self.preview_subframe = TabularDataFrame(self.frame)
+                self.preview_subframe = TabularDataFrame(self.frame, cursor="arrow")
                 self.preview_subframe.set_dataframe(df)
                 self.preview_subframe.set_sort_order(self.node.sort_column or 0, self.node.sort_descending)
                 self.preview_subframe.set_callback(self.preview_changed_callback)
@@ -250,6 +252,7 @@ class ConfiguratorWrapper:
         self.config_scrollbar.set(pos1, pos2)
 
     def choose_plugin(self, plugin_class):
+        print(f"choose {self} {plugin_class}")
         self.node.plugin = plugin_class()
         self.node.prerun(self.logger)
         self.node.is_dirty = True
@@ -259,10 +262,7 @@ class ConfiguratorWrapper:
     def destroy(self):
         if self.config_change_task:
             self.frame.after_cancel(self.config_change_task)
-        if self.config_subframe:
-            self.config_subframe.destroy()
-        if self.preview_subframe:
-            self.preview_subframe.destroy()
+        self.frame.destroy()
 
 
 class ButtonMenu:  # pylint: disable=R0903
@@ -468,6 +468,7 @@ class MainWindow:
             new_config_wrapper = ConfiguratorWrapper(self.main_subframe, node, self.logger, self.node_changed)
             if self.config_wrapper:
                 self.main_subframe.replace_child(self.config_wrapper.frame, new_config_wrapper.frame)
+                self.config_wrapper.destroy()
             else:
                 self.main_subframe.add_child(new_config_wrapper.frame, weight=4)
 
