@@ -22,7 +22,7 @@ class PivotPlugin(PandasProcessPlugin):
     name = "Pivot Tool"
     description = "Groups a dataframe and pivots column values into columns."
     additional = """Expanded column values are duplicated for each combination of pivot values.
-        Missing values default to 0, and duplicate values are summed."""
+        Missing values default to 0, and duplicate values are aggregated."""
     version = VERSION
     link = "https://countess-project.github.io/CountESS/included-plugins/#pivot-tool"
 
@@ -30,17 +30,20 @@ class PivotPlugin(PandasProcessPlugin):
         "columns": PerColumnArrayParam(
             "Columns", ChoiceParam("Role", "Drop", choices=["Index", "Pivot", "Expand", "Drop"])
         ),
+        "aggfunc": ChoiceParam("Aggregation Function", "sum", choices=["sum", "mean", "min", "max"]),
     }
 
     input_columns: Dict[str, np.dtype] = {}
 
-    dataframes: List[pd.DataFrame] = []
+    dataframes: Optional[List[pd.DataFrame]] = None
 
     def prepare(self, sources: List[str], row_limit: Optional[int] = None):
         self.input_columns = {}
+        self.dataframes = []
 
     def process(self, data: pd.DataFrame, source: str, logger: Logger):
         assert isinstance(self.parameters["columns"], PerColumnArrayParam)
+        assert self.dataframes is not None
         self.input_columns.update(get_all_columns(data))
 
         data.reset_index(drop=data.index.names == [None], inplace=True)
@@ -75,7 +78,7 @@ class PivotPlugin(PandasProcessPlugin):
             values=expand_cols,
             index=index_cols,
             columns=pivot_cols,
-            aggfunc="sum",
+            aggfunc=self.parameters["aggfunc"].value,
             fill_value=0,
         )
 
